@@ -1,7 +1,7 @@
 ;;;
 ;;; wiliki/edit - handles edit, preview, and conflict page
 ;;;
-;;;  Copyright (c) 2000-2004 Shiro Kawai, All rights reserved.
+;;;  Copyright (c) 2000-2009  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;  Permission is hereby granted, free of charge, to any person
 ;;;  obtaining a copy of this software and associated documentation
@@ -23,44 +23,56 @@
 ;;;  OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 ;;;  IN THE SOFTWARE.
 ;;;
-;;;  $Id: edit.scm,v 1.14 2004/03/22 11:44:35 shirok Exp $
+;;;  $Id: edit.scm,v 1.22 2007-11-05 22:26:40 shirok Exp $
 ;;;
 
-(select-module wiliki)
+(define-module wiliki.edit
+  (use gauche.parameter)
+  (use srfi-13)
+  (use text.gettext)
+  (use text.diff)
+  (use www.cgi)
+  (use wiliki.core)
+  (use wiliki.log)
+  (use wiliki.page)
+  (use wiliki.macro)
+  (export cmd-edit cmd-preview cmd-commit-edit))
+(select-module wiliki.edit)
 
-(use text.diff)
+(define $$ gettext)
 
-(define (edit-form preview? pagename content mtime logmsg donttouch)
+(define (edit-form preview pagename content mtime logmsg donttouch)
   (define (buttons)
-    (if preview?
-        `((input (@ (type submit) (name preview) (value ,($$ "Preview"))))
-          (input (@ (type submit) (name commit) (value ,($$ "Commit without preview")))))
+    (if preview
         `((input (@ (type submit) (name preview) (value ,($$ "Preview again"))))
-          (input (@ (type submit) (name commit) (value ,($$ "Commit")))))))
+          (input (@ (type submit) (name commit) (value ,($$ "Commit")))))
+        `((input (@ (type submit) (name preview) (value ,($$ "Preview"))))
+          (input (@ (type submit) (name commit) (value ,($$ "Commit without preview")))))))
   (define (donttouch-checkbox)
     `((input (@ (type checkbox) (name donttouch) (value on) (id donttouch)
                 ,@(if donttouch '((checked checked)) '())))
       (label (@ (for donttouch)) ,($$ "Don't update 'Recent Changes'"))))
   
   `((form
-     (@ (method POST) (action ,(cgi-name-of (wiliki))))
-     ,@(buttons) ,@(donttouch-checkbox)
-     (br)
+     (@ (method POST) (action ,(wiliki:url)))
      (input (@ (type hidden) (name c) (value c)))
      (input (@ (type hidden) (name p) (value ,pagename)))
-     (input (@ (type hidden) (name l) (value ,(lang))))
+     (input (@ (type hidden) (name l) (value ,(wiliki:lang))))
      (input (@ (type hidden) (name mtime) (value ,mtime)))
+     ,@(buttons) ,@(donttouch-checkbox)
+     (br)
+     ,@(if preview (list preview '(hr)) '())
      (textarea (@ (name content)
                   (class content)
-                  (rows ,(textarea-rows-of (wiliki)))
-                  (cols ,(textarea-cols-of (wiliki))))
+                  (rows ,(ref (wiliki)'textarea-rows))
+                  (cols ,(ref (wiliki)'textarea-cols)))
                ,content)
      (br)
      (p ,($$ "ChangeLog (brief summary of your edit for later reference):"))
      (textarea (@ (name logmsg)
                   (class logmsg)
                   (rows 2)
-                  (cols ,(textarea-cols-of (wiliki))))
+                  (cols ,(ref (wiliki)'textarea-cols)))
                ,logmsg)
      (br)
      ,@(buttons)
@@ -110,45 +122,58 @@
          It emphasizes a null string, so it's effectively nothing.</p>"))
      )))
 
-(define (cmd-edit pagename)
-  (unless (editable? (wiliki))
+(define (cmd-edit pagename time)
+  (define (get-old-content page)
+    (and-let* ((time)
+               (lines (wiliki-log-recover-content pagename
+                                                  (wiliki:log-file-path (wiliki))
+                                                  (ref page 'content)
+                                                  time)))
+      (string-join lines "\n")))
+  (unless (eq? (ref (wiliki)'editable?) #t)
     (errorf "Can't edit the page ~s: the database is read-only" pagename))
-  (let ((page (wiliki-db-get pagename #t)))
-    (html-page (make <wiliki-page>
-                   :title pagename
-                   :content
-                   (edit-form #t pagename
-                              (ref page 'content)
-                              (ref page 'mtime) "" #f)))))
+  (let* ((page (wiliki:db-get pagename #t))
+         (content (or (get-old-content page) (ref page 'content)))
+         )
+    (wiliki:std-page
+     (make <wiliki-page>
+       :title pagename
+       :content
+       (edit-form #f pagename content (ref page 'mtime) "" #f)))))
 
-(define (cmd-preview pagename content mtime logmsg donttouch)
-  (let ((page (wiliki-db-get pagename #t)))
-    (html-page
+(define (cmd-preview pagename content mtime logmsg donttouch restricted)
+  (let ((page (wiliki:db-get pagename #t)))
+    (wiliki:std-page
      (make <wiliki-page>
        :title (format #f ($$ "Preview of ~a") pagename)
        :content
-       `(,(preview-box (wiliki:format-content content))
-         (hr)
-         ,@(edit-form #f pagename content mtime logmsg donttouch))))))
+       (edit-form (preview-box (wiliki:format-content content))
+                  pagename content mtime logmsg donttouch)))))
 
-(define (cmd-commit-edit pagename content mtime logmsg donttouch)
-  (let ((p   (wiliki-db-get pagename #t))
+;; DONTTOUCH - If #t, don't update RecentChanges.
+;; LIMITED - #t indicates this edit is generated procedurally, like comment
+;;           feature.  It is allowed if EDITABLE? == limited.
+
+(define (cmd-commit-edit pagename content mtime logmsg donttouch limited)
+  (let ((p   (wiliki:db-get pagename #t))
         (now (sys-time)))
 
     (define (erase-page)
       (write-log (wiliki) pagename (ref p 'content) "" now logmsg)
       (set! (ref p 'content) "")
-      (wiliki-db-delete! pagename)
-      (redirect-page (top-page-of (wiliki))))
+      (wiliki:db-delete! pagename)
+      (wiliki:redirect-page (ref (wiliki)'top-page)))
 
     (define (update-page content)
       (when (page-changed? content (ref p 'content))
-        (let1 new-content (expand-writer-macros content)
+        (let1 new-content
+            (parameterize ([wiliki:page-stack (list p)])
+              (expand-writer-macros content))
           (write-log (wiliki) pagename (ref p 'content) new-content now logmsg)
           (set! (ref p 'mtime) now)
           (set! (ref p 'content) new-content)
-          (wiliki-db-put! pagename p :donttouch donttouch)))
-      (redirect-page pagename))
+          (wiliki:db-put! pagename p :donttouch donttouch)))
+      (wiliki:redirect-page pagename))
 
     ;; check if page has been changed.  we should ignore the difference
     ;; of line terminators.
@@ -158,7 +183,7 @@
 
     (define (handle-conflict)
       ;; let's see if we can merge changes
-      (or (and-let* ((logfile (log-file-path (wiliki)))
+      (or (and-let* ((logfile (wiliki:log-file-path (wiliki)))
                      (picked (wiliki-log-pick-from-file pagename logfile)))
             (let ((common (wiliki-log-revert*
                            (wiliki-log-entries-after picked mtime)
@@ -170,7 +195,7 @@
                   (conflict-page p (conflict->diff merged)
                                  content logmsg donttouch)))))
           (if (equal? (ref p 'content) content)
-            (redirect-page pagename) ;; no need to update
+            (wiliki:redirect-page pagename) ;; no need to update
             (let1 diff '()
               (diff-report (ref p 'content) content
                            :writer (lambda (line type)
@@ -187,22 +212,48 @@
             (push! difflist chunk)))
         (reverse! difflist)))
 
+    ;; Ad-hoc filter for mechanical spams.
+    (define (suspicious?)
+      (or
+       ;; Normal wiliki content never includes explicit HTML tags (strictly
+       ;; speaking, the content may have HTML tag within verbatim block.
+       ;; let's see if it becomes a problem or not.
+       (and (string? content) (#/<a\s+href=[\"' ]?\s*http/i content)
+            "literal anchor tag in content")
+       (and (string? logmsg) (#/<a\s+href=[\"' ]?\s*http/i logmsg)
+            "literal anchor tag in logmsg")
+       ;; Some spammer put the same string in content and logmsg.
+       (and (not (equal? content "")) (equal? content logmsg)
+            "content and logmsg are the same")
+       ;; Check blacklist.
+       (and (wiliki:contains-spam? content)
+            "url-hit-blacklist")))
+
     ;; The body of cmd-commit-edit
     ;; If content is empty and the page is not the top page, we erase
     ;; the page.
-    (unless (editable? (wiliki))
-      (errorf "Can't edit the page ~s: the database is read-only" pagename))
-    (if (or (not (ref p 'mtime)) (eqv? (ref p 'mtime) mtime))
-      (if (and (not (equal? pagename (top-page-of (wiliki))))
+    (let1 editable (ref (wiliki)'editable?)
+      (when (or (not editable)
+                (and (not limited) (eq? editable 'limited)))
+        (errorf "Can't edit the page ~s: the database is read-only" pagename)))
+    (cond
+     [(suspicious?)
+      => (lambda (reason)
+           (wiliki:log-event "rejecting spam on ~s (~a): content=~s logmsg=~s"
+                             pagename reason content logmsg)
+           (wiliki:redirect-page (ref (wiliki)'top-page)))]
+     [(or (not (ref p 'mtime)) (eqv? (ref p 'mtime) mtime))
+      (if (and (not (equal? pagename (ref (wiliki)'top-page)))
                (string-every #[\s] content))
         (erase-page)
-        (update-page content))
-      (handle-conflict))))
+        (update-page content))]
+     [else (handle-conflict)])
+    ))
 
 (define (conflict-page page diff content logmsg donttouch)
-  (html-page
+  (wiliki:std-page
    (make <wiliki-page>
-     :title (string-append (title-of (wiliki))": "($$ "Update Conflict"))
+     :title (string-append (ref (wiliki)'title)": "($$ "Update Conflict"))
      :content
      `((stree ,($$ "<p>It seems that somebody has updated this page
        while you're editing.  The difference is snown below.
@@ -216,7 +267,7 @@
        ,(wiliki:format-diff-pre diff)
        (a (@ (name "edit")) (hr))
        ,($$ "<p>The following shows what you are about to submit.  Please re-edit the content and submit again.</p>")
-       ,@(edit-form #t (ref page 'key) content (ref page 'mtime) logmsg donttouch)
+       ,@(edit-form #f (ref page 'key) content (ref page 'mtime) logmsg donttouch)
        ))))
 
 (define (preview-box content)
@@ -225,5 +276,20 @@
     (tr (td (@ (class "preview"))
             ,@content))))
 
+;; NB: we assume write-log is always called during the main database
+;; is locked, so we don't do any locking here.
+(define (write-log wiliki pagename old new timestamp logmsg)
+  (and-let* ((logfile (wiliki:log-file-path wiliki)))
+    (let1 content (wiliki-log-create
+                   pagename new old
+                   :timestamp timestamp
+                   :remote-addr (or (cgi-get-metavariable "REMOTE_ADDR") "")
+                   :remote-user (or (cgi-get-metavariable "REMOTE_USER") "")
+                   :message (or logmsg ""))
+      (call-with-output-file logfile
+        (lambda (p) (display content p) (flush p))
+        :if-exists :append))
+    ))
+
 (provide "wiliki/edit")
 
